@@ -1,11 +1,10 @@
 """Module that contains the parser for PDDL+ domain files."""
 import logging
-import types
 from pathlib import Path
-from typing import List, Dict, Union, NoReturn, Tuple
+from typing import List, Dict, Union, Tuple
 
 from pddl_plus_parser.models import Domain, PDDLType, Predicate, PDDLConstant, PDDLFunction, Action, SignatureType, \
-    NumericalExpressionTree, construct_expression_tree, ConditionalEffect
+    NumericalExpressionTree, construct_expression_tree, ConditionalEffect, UniversalQuantifiedEffect
 from .parsing_utils import parse_signature
 from .pddl_tokenizer import PDDLTokenizer
 
@@ -22,13 +21,13 @@ class DomainParser:
     tokenizer: PDDLTokenizer
     logger: logging.Logger
     partial_parsing: bool
-    enable_conjunctions: bool
+    enable_disjunctions: bool
 
-    def __init__(self, domain_path: Path, partial_parsing: bool = False, enable_conjunctions: bool = False):
+    def __init__(self, domain_path: Path, partial_parsing: bool = False, enable_disjunctions: bool = False):
         self.tokenizer = PDDLTokenizer(domain_path)
         self.logger = logging.getLogger(__name__)
         self.partial_parsing = partial_parsing
-        self.enable_conjunctions = enable_conjunctions
+        self.enable_disjunctions = enable_disjunctions
 
     def parse_types(self, types: List[str]) -> Dict[str, PDDLType]:
         """Parse the types of the domain.
@@ -179,7 +178,7 @@ class DomainParser:
 
     def parse_disjunctive_numeric_preconditions(
             self, numeric_preconditions_node: List[Union[str, List[str]]],
-            new_action: Action, domain_functions: Dict[str, PDDLFunction]) -> NoReturn:
+            new_action: Action, domain_functions: Dict[str, PDDLFunction]) -> None:
         """Parse a set of disjunctive numeric preconditions.
 
         :param numeric_preconditions_node: the node that contains the numeric preconditions.
@@ -199,7 +198,7 @@ class DomainParser:
     def parse_preconditions(self, preconditions_ast: List[Union[str, List[str]]], new_action: Action,
                             domain_functions: Dict[str, PDDLFunction],
                             domain_predicates: Dict[str, Predicate],
-                            domain_constants: Dict[str, PDDLConstant]) -> NoReturn:
+                            domain_constants: Dict[str, PDDLConstant]) -> None:
         """Parse the preconditions of a single action.
 
         :param preconditions_ast: the AST representation of the action's preconditions.
@@ -217,7 +216,7 @@ class DomainParser:
             self.logger.warning("Received an action with no preconditions.")
             return
 
-        if preconditions_ast[0] != "and" and not self.enable_conjunctions:
+        if preconditions_ast[0] != "and" and not self.enable_disjunctions and len(preconditions_ast[1:]) > 1:
             raise SyntaxError(f"Only accepting conjunctive preconditions! Action - {new_action.name} does not conform!")
 
         for precondition_node in preconditions_ast[1:]:
@@ -257,7 +256,7 @@ class DomainParser:
                                   positive_conditionals: List[Predicate], negative_conditionals: List[Predicate],
                                   numeric_conditionals: List[NumericalExpressionTree],
                                   signature: Dict[str, PDDLType], domain_functions: Dict[str, PDDLFunction],
-                                  domain_constants: Dict[str, PDDLConstant]) -> NoReturn:
+                                  domain_constants: Dict[str, PDDLConstant]) -> None:
         """Parse a single conditional from the conditional effect.
 
         :param conditional: the single conditional effect to parse.
@@ -282,14 +281,13 @@ class DomainParser:
         else:
             positive_conditionals.append(self.parse_untyped_predicate(conditional, signature, domain_constants))
 
-    def _parse_conditionals(
-            self, conditional_ast: List[Union[str, List[str]]],
-            signature: Dict[str, PDDLType], domain_functions: Dict[str, PDDLFunction],
-            domain_constants: Dict[str, PDDLConstant]) -> Tuple[
+    def _parse_antecedents(self, antecedent_ast: List[Union[str, List[str]]],
+                           signature: Dict[str, PDDLType], domain_functions: Dict[str, PDDLFunction],
+                           domain_constants: Dict[str, PDDLConstant]) -> Tuple[
         List[Predicate], List[Predicate], List[NumericalExpressionTree]]:
-        """Parse the conditionals for a conditional effect of an action.
+        """Parse the antecedents for a conditional effect of an action.
 
-        :param conditional_ast: the AST representation of the conditionals.
+        :param antecedent_ast: the AST representation of the conditionals.
         :param signature: the action's signature.
         :param domain_functions: the functions that exist in the domain.
         :param domain_constants: the constants that exist in the domain.
@@ -298,20 +296,20 @@ class DomainParser:
         positive_conditionals = []
         negative_conditionals = []
         numeric_conditionals = []
-        if conditional_ast[0] == "and":
+        if antecedent_ast[0] == "and":
             self.logger.debug("Received a conjunctive conditional effect with more than one condition.")
-            for conditional in conditional_ast[1:]:
+            for conditional in antecedent_ast[1:]:
                 self._parse_single_conditional(conditional, positive_conditionals, negative_conditionals,
                                                numeric_conditionals, signature, domain_functions, domain_constants)
 
         else:
-            self._parse_single_conditional(conditional_ast, positive_conditionals, negative_conditionals,
+            self._parse_single_conditional(antecedent_ast, positive_conditionals, negative_conditionals,
                                            numeric_conditionals, signature, domain_functions, domain_constants)
 
         return positive_conditionals, negative_conditionals, numeric_conditionals
 
     def _parse_single_conditional_effect(self, action, conditional_effect_ast, add_effects, del_effects,
-                                         numeric_effects, domain_constants, domain_functions) -> NoReturn:
+                                         numeric_effects, domain_constants, domain_functions) -> None:
         """
 
         :param action: the action that is being parsed.
@@ -334,11 +332,14 @@ class DomainParser:
         else:
             add_effects.append(self.parse_untyped_predicate(conditional_effect_ast, action.signature, domain_constants))
 
-    def _construct_conditional_effects(self, conditional_effect_ast: List[Union[str, List[str]]],
-                                       positive_conditionals: List[Predicate], negative_conditionals: List[Predicate],
-                                       numeric_conditionals: List[NumericalExpressionTree],
-                                       action: Action, domain_functions: Dict[str, PDDLFunction],
-                                       domain_constants: Dict[str, PDDLConstant]) -> NoReturn:
+    def _construct_conditional_effects(
+            self, conditional_effect_ast: List[Union[str, List[str]]],
+            positive_conditionals: List[Predicate],
+            negative_conditionals: List[Predicate],
+            numeric_conditionals: List[NumericalExpressionTree],
+            action: Action,
+            domain_functions: Dict[str, PDDLFunction],
+            domain_constants: Dict[str, PDDLConstant]) -> ConditionalEffect:
         """Parse all the conditional effects that are under the same when statement - can be composite.
 
         :param conditional_effect_ast: the AST representation of the conditional effects.
@@ -368,16 +369,102 @@ class DomainParser:
         conditional_effect.positive_conditions = set(positive_conditionals)
         conditional_effect.negative_conditions = set(negative_conditionals)
         conditional_effect.numeric_conditions = set(numeric_conditionals)
+        return conditional_effect
+
+    def _extend_action_to_contain_quantified_objects(
+            self, action: Action, quantified_parameter_name: str, quantified_type: PDDLType) -> Action:
+        """Extends the relevant parts of the action to include the quantified object.
+
+        :param action: the action to extend.
+        :param quantified_parameter_name: the name of the parameter that is added to the extended action.
+        :param quantified_type: the type of the quatified object.
+        :return: the extended action.
+        """
+        combined_signature = {**action.signature, quantified_parameter_name: quantified_type}
+        extended_action = Action()
+        extended_action.name = action.name
+        extended_action.signature = combined_signature
+        return extended_action
+
+    def parse_conditional_effect(
+            self, conditional_effect_ast: List[Union[str, List[str]]], action: Action,
+            domain_functions: Dict[str, PDDLFunction],
+            domain_constants: Dict[str, PDDLConstant]) -> None:
+        """Parse a conditional effect of an action.
+
+        :param conditional_effect_ast: the AST representation of the conditional effect.
+        :param action: the action that is being parsed.
+        :param domain_functions: the functions that exist in the domain.
+        :param domain_constants: the constants that exist in the domain.
+        """
+        self.logger.debug("Parsing conditional effect node.")
+        if len(conditional_effect_ast[1:]) != 2:
+            raise SyntaxError(f"Conditional effect scheme does not match for action {action.name}!")
+
+        positive_conditionals, negative_conditionals, numeric_conditionals = self._parse_antecedents(
+            conditional_effect_ast[1], action.signature, domain_functions, domain_constants)
+        conditional_effect = self._construct_conditional_effects(conditional_effect_ast[2], positive_conditionals,
+                                                                 negative_conditionals,
+                                                                 numeric_conditionals, action, domain_functions,
+                                                                 domain_constants)
         action.conditional_effects.add(conditional_effect)
 
-    def parse_effects(self, effects_ast: List[Union[str, List[str]]], new_action: Action,
+    def parse_universally_quantified_effect(
+            self, universal_quantifier_ast: List[Union[str, List[str]]],
+            action: Action,
+            domain_types: Dict[str, PDDLType],
+            domain_functions: Dict[str, PDDLFunction],
+            domain_constants: Dict[str, PDDLConstant]) -> None:
+        """Parse a universally quantified effect of an action.
+
+        :param universal_quantifier_ast: the ast representation of the universally quantified effect.
+        :param action: the action that is being parsed.
+        :param domain_types: the types that exist in the domain.
+        :param domain_functions: the functions that exist in the domain.
+        :param domain_constants: the constants that exist in the domain.
+        """
+        self.logger.debug("Parsing universally quantified effect node.")
+        if len(universal_quantifier_ast) != 2:
+            raise SyntaxError(f"Universal quantifier scheme does not match for action {action.name}!")
+
+        quantified_object_components = universal_quantifier_ast[0]
+        # quantified object components are of the form ( ?x - type )
+        if len(quantified_object_components) != 3:
+            raise SyntaxError(f"Quantified object scheme does not match for action {action.name}!"
+                              f"Expected 3 components, got {len(quantified_object_components)}!")
+
+        quantified_parameter_name = quantified_object_components[0]
+        quantified_type = domain_types[quantified_object_components[2]]
+        extended_action = self._extend_action_to_contain_quantified_objects(
+            action, quantified_parameter_name, quantified_type)
+
+        conditional_effect_ast = universal_quantifier_ast[1]
+        positive_conditionals, negative_conditionals, numeric_conditionals = self._parse_antecedents(
+            conditional_effect_ast[1], extended_action.signature, domain_functions, domain_constants)
+        conditional_effect = self._construct_conditional_effects(conditional_effect_ast[2],
+                                                                 positive_conditionals,
+                                                                 negative_conditionals,
+                                                                 numeric_conditionals,
+                                                                 extended_action,
+                                                                 domain_functions,
+                                                                 domain_constants)
+
+        universal_quantifier = UniversalQuantifiedEffect(quantified_parameter=quantified_parameter_name,
+                                                         quantified_type=quantified_type,
+                                                         conditional_effect=conditional_effect)
+        action.universal_effects.add(universal_quantifier)
+
+    def parse_effects(self, effects_ast: List[Union[str, List[str]]],
+                      new_action: Action,
+                      domain_types: Dict[str, PDDLType],
                       domain_functions: Dict[str, PDDLFunction],
                       domain_predicates: Dict[str, Predicate],
-                      domain_constants: Dict[str, PDDLConstant]) -> NoReturn:
+                      domain_constants: Dict[str, PDDLConstant]) -> None:
         """Parse the effects of a single action.
 
         :param effects_ast: the AST representation of the action's effects.
         :param new_action: the action that is currently being parsed.
+        :param domain_types: the types that exist in the domain.
         :param domain_functions: the functions that exist in the domain.
         :param domain_predicates: the predicates that are defined in the domain.
         :param domain_constants: the domains that might exist in the domain.
@@ -401,18 +488,13 @@ class DomainParser:
                     self.parse_untyped_predicate(effect_node[1], new_action.signature, domain_constants))
                 continue
 
+            if effect_node[0] == "forall":
+                self.parse_universally_quantified_effect(effect_node[1:], new_action, domain_types, domain_functions,
+                                                         domain_constants)
+                continue
+
             if effect_node[0] == "when":
-                if len(effect_node[1:]) != 2:
-                    raise SyntaxError(f"Conditional effect scheme does not match for action {new_action.name}!")
-
-                self.logger.debug("Found a conditional effect!")
-                positive_conditionals, negative_conditionals, numerical_conditionals = \
-                    self._parse_conditionals(effect_node[1], new_action.signature, domain_functions, domain_constants)
-
-                self._construct_conditional_effects(effect_node[2], positive_conditionals, negative_conditionals,
-                                                    numerical_conditionals, new_action, domain_functions,
-                                                    domain_constants)
-
+                self.parse_conditional_effect(effect_node, new_action, domain_functions, domain_constants)
                 continue
 
             if effect_node[0] in ASSIGNMENT_OPS:
@@ -457,7 +539,7 @@ class DomainParser:
 
             if action_label_item == ":effect" and not self.partial_parsing:
                 self.logger.debug(f"Starting to parse the effects of the action - {new_action.name}")
-                self.parse_effects(next(action_section_iterator), new_action, domain_functions,
+                self.parse_effects(next(action_section_iterator), new_action, domain_types, domain_functions,
                                    domain_predicates, domain_constants)
 
         return new_action
@@ -501,3 +583,5 @@ class DomainParser:
                 # TODO: complete once I finish numeric actions support
 
         return domain
+
+
